/*
 * Simple GLUT-based pixel editor for calibrating projector color response.
 * Stuart Levy, NCSA, University of Illinois, 2001
 * slevy@ncsa.uiuc.edu
 */

#include <GL/glut.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

int nx, ny;
int cx, cy, cz;
int redo;
int blinkon;
int doall;
int changed, everchanged;
char *curfname;
unsigned char *rgb, *oldrgb;
unsigned char copybuf[3];

void reshaped(int w, int h)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho( 0, nx, 0, ny, -1, 1 );

    glMatrixMode(GL_MODELVIEW);
    glViewport(0, 0, w, h);
    glLoadIdentity();
}

void predo() {
    memcpy(oldrgb, rgb, 3*nx*ny);
}
    
void blinker( int on ) {
    blinkon = !blinkon || on;
    if(!on) glutTimerFunc( 500, blinker, 0 );
    glutPostRedisplay();
}

unsigned char *cell( int x, int y, int z ) {
    return &rgb[ ((ny-1-y)*nx + x) * 3 + z ];
}

void draw() {
    int x, y;

    glClearColor( 0,0,0,1 );
    glClear( GL_COLOR_BUFFER_BIT );
    glBegin(GL_QUADS);
    for(y = 0; y < ny; y++) {
	for(x = 0; x < nx; x++) {
	    glColor3ubv( cell(x,y,0) );
	    glVertex2i( x, y );
	    glVertex2i( x, y+1 );
	    glVertex2i( x+1, y+1 );
	    glVertex2i( x+1, y );
	}
    }
    glEnd();
    if(blinkon) {
	unsigned char *cp = cell(cx,cy,0);
	glColor3ub( cp[0]^0x80, cp[1]^0x80, cp[2]^0x80 );
	glLineWidth( 2 );
	glBegin( GL_LINE_LOOP );
	glVertex2f( cx+.01, cy+.01 );
	glVertex2f( cx+.01, cy+.99 );
	glVertex2f( cx+.99, cy+.99 );
	glVertex2f( cx+.99, cy+.01 );
	glEnd();
	if(doall) {
	    int x, y;
	    glLineWidth( 1 );
	    glBegin( GL_LINES );
	    for(y = 0; y <= ny; y++) {
		glVertex2i( 0, y );
		glVertex2i( nx, y );
	    }
	    for(x = 0; x <= nx; x++) {
		glVertex2i( x, 0 );
		glVertex2i( x, ny );
	    }
	    glEnd();
	}
    }
    glutSwapBuffers();
}

int getint( FILE *f ) {
    int c, v;
    while(isspace(c = getc(f)) || c == '#') {
	if(c == '#') {
	    while((c = getc(f)) != EOF && c != '\n')
		;
	}
    }
    if(c == EOF)
	return -1;
    if(isdigit(c)) {
	v = c - '0';
	while(isdigit(c = getc(f))) {
	    v = v*10 + c - '0';
	}
	if(c != EOF)
	    ungetc(c, f);
	return v;
    }
    return -1;
}

void init( char *fname ) {
    FILE *f = fopen(fname, "rb");
    char line[512], *cp;
    int c, maxval;
    if(f == NULL) {
	fprintf(stderr, "%s: cannot open input\n", fname);
	exit(1);
    }
    c = getc(f);
    if(c != 'P' || getc(f) != '6') {
	fprintf(stderr, "%s: not a binary PPM image file\n", fname);
	exit(1);
    }
    nx = getint(f);
    ny = getint(f);
    maxval = getint(f);
    if(maxval != 255) {
	fprintf(stderr, "%s: not a binary PPM image file\n", fname);
	exit(1);
    }
    while((c = getc(f)) != EOF && c != '\n')
	;
    rgb = (unsigned char *)malloc( nx*ny*3 );
    oldrgb = (unsigned char *)malloc( nx*ny*3 );
    if(fread( rgb, nx*ny*3, 1, f ) <= 0) {
	fprintf(stderr, "%s: couldn't read all expected %d*%d*3 image bytes!\n",
		fname, nx, ny);
	exit(1);
    }
    cp = strrchr( fname, '/' );
    curfname = cp ? cp+1 : fname;	/* like `basename inputfile` */
}

void outnit( char *fname ) {
    time_t now;
    char newfname[512];
    FILE *f;
    int i;
    sprintf(newfname, "%.500s.cal", fname);
    for(i = 0; i < 10; i++) {
	f = fopen(newfname, "wb");
	if(f == NULL) {
	    fprintf(stderr, "%s: Can't create output\n", newfname);
	    strcat(newfname, "+");
	}
    }
    if(f == NULL)
	return;
    fprintf(f, "P6 #  Calibrated version of %s\n", fname);
    time(&now);
    fprintf(f, "%d %d  # on %s", nx, ny, ctime(&now));
    fprintf(f, "255\n");
    fwrite( rgb, nx*ny*3, 1, f );
    fclose(f);
    fprintf(stderr, "Wrote to %s\n", newfname);
}

void report() {
    unsigned char *cp = cell(cx,cy,0);
    printf( "\r[%2d %2d] r%3d g%3d b%3d  ", cx, cy, cp[0],cp[1],cp[2]);
    fflush(stdout);
}
    
int clamp(int val) {
    return val<0 ? 0 : val>255 ? 255 : val;
}

void change( int val, int asoffset ) {
    int v;
    unsigned char *cp;

    if(!changed) {
	predo();
	changed = 1;
	everchanged = 1;
    }
    cp = cell(cx,cy,cz);
    if(doall) {
	int x, y;
	for(y = 0; y < ny; y++) {
	    for(x = 0; x < nx; x++) {
		cp = cell( x, y, cz );
		*cp = clamp( asoffset ? *cp + val : val );
	    }
	}
    } else {
	*cp = clamp( asoffset ? *cp + val : val );
    }
}

void undo() {
    unsigned char *t;
    if(changed) {
	t = rgb;
	rgb = oldrgb;
	oldrgb = rgb;
    }
}

void keybo(unsigned char key, int x, int y)
{
    int i;
    switch(key) {
    case 'w':  outnit(curfname); everchanged = changed = 0; report(); break;
    case 033:  if(everchanged) outnit(curfname); exit(0);
    case 'Q':  exit(0);
    case 'r':  cz = 0; change( 1, 1 ); report(); blinker(1); break;
    case 'R':  cz = 0; change( -1, 1 ); report(); blinker(1); break;
    case 'g':  cz = 1; change( 1, 1 ); report(); blinker(1); break;
    case 'G':  cz = 1; change( -1, 1 ); report(); blinker(1); break;
    case 'b':  cz = 2; change( 1, 1 ); report(); blinker(1); break;
    case 'B':  cz = 2; change( -1, 1 ); report(); blinker(1); break;
    case 'c':	memcpy( copybuf, cell(cx,cy,cz), 3 ); report();
		printf("copied [v to paste]\n");
		report();
		break;
    case 'v':	for(cz=0; cz<3; cz++) change( copybuf[0], 0 );
		report(); printf("pasted"); fflush(stdout);
		break;
    case 'u':	undo(); break;
    case '.':   doall = 0; break;
    case '*':   doall = 1; break;
    case '\r': printf("\n"); report(); break;
    case '<':   {
		char fname[128], *ep;
		  printf("Load file: "); fflush(stdout);
		  fgets(fname, sizeof(fname), stdin);
		  ep = fname+strlen(fname);
		  while(--ep >= fname && *ep <= ' ') *ep = '\0';
		  init( fname );
		}
		break;
		  
    }
    glutPostRedisplay();
}

void specialkeybo(int key, int x, int y)
{
    switch (key) {
    case GLUT_KEY_UP:		if(cy < ny-1) cy++; break;
    case GLUT_KEY_DOWN:		if(cy > 0) cy--; break;
    case GLUT_KEY_RIGHT:	if(cx < nx-1) cx++; break;
    case GLUT_KEY_LEFT:		if(cx > 0) cx--; break;
    default: return;
    }
    doall = 0;
    changed = 0;
    blinker(1);
    report();
}


int main(int argc, char *argv[])
{
    if(argc <= 1) {
	fprintf(stderr, "Usage: tilecal file.ppm\n\
Use arrow keys to step through image, r/R/g/G/b/B to brighten/darken tiles,\n\
w to write edited image to file.ppm.cal, ESC to quit.\n");
	exit(1);
    }

    init(argv[1]);

    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
    glutCreateWindow("tile calibration");
    glutDisplayFunc(draw);
    glutKeyboardFunc(keybo);
    glutSpecialFunc(specialkeybo);
    glutReshapeFunc(reshaped);
    glutTimerFunc( 500, blinker, 0 );
    glutMainLoop();
    return 0;
}
